#include <QWebFrame>
#include <QCloseEvent>
#include "mapviewdlg.h"
#include "ui_mapviewdlg.h"


// ************************************************
// Constructor function.
// ************************************************
MapViewDlg::MapViewDlg(app_config *Config, QWidget *parent) : QDialog(parent),  ui(new Ui::MapViewDlg)
{
   // Set up the user interface.
   ui->setupUi(this);

   // Save pointer to 'Config' structure.
   config = Config;

   // Add items to 'Zoom Level' combo-box and set default zoom level.
   ui->ZoomLevelBox->addItem("Auto");
   for (int i=0;i<22;i++) {
      QString ZL = QString::number(i);
      ui->ZoomLevelBox->addItem(ZL);
      if (config->DefaultZoomLevel == ZL) ui->ZoomLevelBox->setCurrentIndex(i+1);
   }

   // Add items to 'Map Type' combo-box and set default map type.
   ui->MapTypeBox->addItem("Roadmap", "roadmap");
   ui->MapTypeBox->addItem("Satellite", "satellite");
   ui->MapTypeBox->addItem("Terrain", "terrain");
   ui->MapTypeBox->addItem("Hybrid", "hybrid");
   if (config->DefaultMapType == "satellite") ui->MapTypeBox->setCurrentIndex(1);
   if (config->DefaultMapType == "terrain") ui->MapTypeBox->setCurrentIndex(2);
   if (config->DefaultMapType == "hybrid") ui->MapTypeBox->setCurrentIndex(3);
}




// ************************************************
// Destructor function.
// ************************************************
MapViewDlg::~MapViewDlg()
{
   delete ui;
}




// ************************************************
// Show a map with markers for the given locations.
// ************************************************
void MapViewDlg::ShowMap(QListCoord *loclist)
{
   // Save the list of locations.
   LocList = *loclist;

   // Show window if it is hidden.
   if (this->isHidden()) {
      // Restore last used size/position of window.
      this->move(WindowPos);
      // Show window.
      this->setVisible(true);
   }

   // Show the map.
   LoadMap();
}




// *************************************************
// User pressed the 'Close' button. Hide the window.
// *************************************************
void MapViewDlg::CloseButton()
{
   // Save size/position of map view.
   WindowPos = this->pos();

   // Hide the map view.
   this->setVisible(false);
}




// *********************************************
// Load the map from Google server.
// *********************************************
void MapViewDlg::LoadMap()
{
   // Abort if window is hidden.
   if (this->isHidden()) return;

   // Create the URL for the web page.
   BuildURL();

   // Show the map.
   QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
   ui->MapView->load(QUrl(URL));
   QApplication::restoreOverrideCursor();
}




// *********************************************
// User has selected a zoom level.
// *********************************************
void MapViewDlg::ZoomLevelBox(QString value)
{
   // Set new zoom level.
   zoomlevel = value;
   ZoomLevel = (value != "Auto");

   // Update the map.
   LoadMap();
}




// *********************************************
// User has selected a map type.
// *********************************************
void MapViewDlg::MapTypeBox()
{
   // Get selected map type.
   maptype = ui->MapTypeBox->itemData(ui->MapTypeBox->currentIndex()).toString();
   MapType = (maptype != "roadmap");

   // Update the map.
   LoadMap();
}




// **********************************************************
// Build the URL for the map. Google Static Maps API is used.
// Locations on the map will be skipped if URL gets too long.
// **********************************************************
void MapViewDlg::BuildURL()
{
   QString Locat;

   // Build the first static part of the URL.
   URL = "http://maps.googleapis.com/maps/api/staticmap";

   // Set size and scale of map image.
   URL.append("?size=640x640&scale=2");

   // Set type of map.
   if (MapType) URL.append("&maptype=" + maptype);

   // Set zoom level of map.
   if (ZoomLevel) URL.append("&zoom=" + zoomlevel);

   // Add API key.
   if (config->StaticMapApiKey.size()) URL.append("&key=" + config->StaticMapApiKey);

   // Set sensor to false.
   URL.append("&sensor=false");

   // Create markers.
   URL.append("&markers=size:mid|color:yellow");
   for (int i=0;i<LocList.size();i++) {
      // Set marker locations.
      Locat = "|";
      Locat.append(QString::number(LocList.at(i).lat, 'f', 6));
      Locat.append(",");
      Locat.append(QString::number(LocList.at(i).lon, 'f', 6));
      if (URL.size() + Locat.size() <= 1880) URL.append(Locat);
   }
}




// *********************************************
// The dialog window is about to be closed.
// This event is triggered by CleanUpOnExit().
// *********************************************
void MapViewDlg::closeEvent(QCloseEvent *event)
{
    // Save current zoom level to configuration.
    config->DefaultZoomLevel = ui->ZoomLevelBox->currentText();

    // Save current map type to configuration.
    config->DefaultMapType = ui->MapTypeBox->itemData(ui->MapTypeBox->currentIndex()).toString();

    // Accept the close event.
    event->accept();
}

