#include <QDebug>
#include "fortran.h"
#include "library.h"
#include "readsfiles.h"


// **********************************************************
// This function is used to initialize the class with
// a full list of all s-files in the seisan database.
// This function must be called before read is started.
// **********************************************************
void readsfiles::Init(QStringList* SfileList, QMutex *mtx_db_cblock)
{
   sfilelist = SfileList;
   mtx_common_block = mtx_db_cblock;
}




// **********************************************************
// Reads all S-files into database-nodes and adds
// them to the database using the AddDbEvent signal.
// **********************************************************
void readsfiles::run()
{
   QDate fDate;
   bool FileProblem;
   event_node_ *Node;
   char file[81], status;
   int num_files, filenr, progress;

   // Reset the cancel flag.
   mtx_terminate.lock();
   terminate = false;
   mtx_terminate.unlock();

   // Initialize variables.
   progress = 0;
   rea4_.rea_err_unit = 6;
   FileProblem = false;
   num_files = sfilelist->size();

   // Read S-files into nodes and add nodes to database.
   // Operation will stop if we run out of memory, or
   // if we recieve a request to terminate the operation.
   for (filenr=0;filenr<num_files;filenr++) {
      // Read S-file into memory.
      strcpy(file, sfilelist->at(filenr).toLatin1().constData());
      mtx_common_block->lock();
      Node = ReadSfile(file, &status);
      mtx_common_block->unlock();

      // If file was sucessfully read, add node to database.
      // Files with invalid dates will not be added.
      if (Node) emit AddDbEvent(Node);

      // Handle warnings and errors.
      if (status) {
         if (status == 1) {
            // The file could not be opened.
            FileProblem = true;
            Message = "Unable to open event file '";
            Message.append(file); Message.append("'.");
            emit WriteLog(logmodel::error, Message, sfilelist->at(filenr));
         }
         if (status == 2) {
            // The current file has errors and cannot be read.
            // Write an error message in the program log.
            FileProblem = true;
            Message = "An error was detected while reading event file '";
            Message.append(file); Message.append("'. File has been skipped.");
            emit WriteLog(logmodel::error, Message, sfilelist->at(filenr));
         }
         if (status == 3) {
            // We are out of memory, reading will stop here.
            // Signal database class that reading is done.
            emit ReadSfilesDone(1);
            return;
         }
         if (status == 4) {
            // The current file has some non-fatal problem, but has
            // been read. Write a warning message in the program log.
            Message = "A problem was detected while reading event file '";
            Message.append(file); Message.append("'.");
            emit WriteLog(logmodel::warning, Message, sfilelist->at(filenr));
         }
         if (status == 5) {
            // The current event file has an invalid date.
            // Write an error message in the program log.
            FileProblem = true;
            Message = "Event file has invalid date: '";
            Message.append(file); Message.append("'.");
            emit WriteLog(logmodel::error, Message, sfilelist->at(filenr));
         }
      }

      // For every 300 file, check if thread has been asked to terminate.
      if (++progress > 300) {
         // Check if thread has been asked to terminate.
         mtx_terminate.lock();
         if (terminate) {
            mtx_terminate.unlock();
            emit ReadSfilesDone(2);
            return;
         }
         progress = 0;
         mtx_terminate.unlock();
      }
   }

   // Signal database class that reading is done.
   if (FileProblem) emit ReadSfilesDone(3);
   if (!FileProblem) emit ReadSfilesDone(0);
}




// **********************************************************
// This function can be called to stop the thread function.
// Thread will stop at the earliest convenience.
// **********************************************************
void readsfiles::Cancel()
{
   // Check if thread is running.
   if (isRunning()) {
      // Raise the terminate flag.
      mtx_terminate.lock();
      terminate = true;
      mtx_terminate.unlock();
   }
}

