#include <QtGlobal>
#include <QDate>
#include <QString>
#include <QStringList>
#include <stdlib.h>

#ifndef TYPES_H
#define TYPES_H

// Type definitions.
typedef QList<int> QListInt;
typedef QList<struct event_id> QListEventID;
typedef QList<struct coord> QListCoord;

// Program limitations.
#define MAX_EVENTS  1000000             // Max number of S-files the program can handle. Must be divisible by 1000.
#define MAX_SFILE_LINES  10000          // Max number of lines in an S-file.
#ifdef Q_OS_LINUX
#define MAX_PATH 260                    // Max allowed length of full path to a file.
#endif

// Define name of files.
#define CONSOLE_FILE (const char *)("se-console.out")
#define WORKDIR_LOCK_FILE (const char *)("se-workdir.lock")

// Defines related to Google Earth.
#define PRI_KML_FILE (const char *)("se-gmap.kml")
#define SEC_KML_FILE (const char *)("gmap.cur.kml")

// Defines related to the use of QMessageBox.
#define QMB_YES QMessageBox::Yes
#define QMB_NO QMessageBox::No



//**********************************************
// Define names for all event list view columns.
//**********************************************
#define NumElvColumns  27
#define ColRowNumber    0
#define ColAction       1
#define ColDateTime     2
#define ColLatitude     3
#define ColLongitude    4
#define ColDepth        5
#define ColModel        6
#define ColAgency       7
#define ColRMS          8
#define ColGap          9
#define ColErrLat       10
#define ColErrLon       11
#define ColErrDep       12
#define ColDist         13
#define ColType         14
#define ColMInt         15
#define ColNrSta        16
#define ColM            17
#define ColMW           18
#define ColML           19
#define ColMC           20
#define ColMb           21
#define ColMB           22
#define ColMs           23
#define ColMS           24
#define ColLocality     25
#define ColSFile        26


enum sortCol {
   RowNr,                              // Cannot sort by this column.
   Action,                             // Always present in event node.
   DateTime,                           // Not always present, or partly present.
   Latitude,                           // Not always present in event node.
   Longitude,                          // Not always present in event node.
   Depth,                              // Not always present in event node.
   Model,                              // Not always present in event node.
   Agency,                             // Always present in event node.
   RMS,                                // Not always present in event node.
   Gap,                                // Not always present in event node.
   ErrLat,                             // Not always present in event node.
   ErrLon,                             // Not always present in event node.
   ErrDep,                             // Not always present in event node.
   Dist,                               // Always present in event node.
   Type,                               // Not always present in event node.
   MInt,                               // Not always present in event node.
   NSta,                               // Not always present in event node.
   M,                                  // Not always present in event node.
   MW,                                 // Not always present in event node.
   ML,                                 // Not always present in event node.
   MC,                                 // Not always present in event node.
   Mb,                                 // Not always present in event node.
   MB,                                 // Not always present in event node.
   Ms,                                 // Not always present in event node.
   MS,                                 // Not always present in event node.
   Locality,                           // Not always present in event node.
   SFile                               // Always present in event node.
};


enum app_status {
   db_closed,
   db_loading,
   db_open
};


struct event_id {
   int index;
   int id;
};


struct db_filter {
   QString Description;
   QString InclQuery;
   QString ExclQuery;
};


struct coord {
   float lat;
   float lon;
};


// ****************************************
// Class holding the program configuration.
// ****************************************
struct app_config {
    // Time interval dialog.
    QDate StartDate;                 // Start of interval (oldest date).
    QDate EndDate;                   // End of interval (newest date).
    char EndDateType;                // Indicates one of 1='Current date', 2='End of month', or 3='User specified'.

    // Configuration dialog.
    int TimeInterval;                // Default timespan between start date and stop date (in days).
    bool AutoLoad;                   // Specifies if Explorer will load last used database at startup.

    // Database.
    int LastUsedDatabaseType;        // Type of database last opened. Valus of 0 means 'no last used database'.
    QString LastUsedDatabaseName;    // Name of last used database.
    QString LastUsedDatabasePath;    // Path to last used database.

    // Event List View.
    bool EvtListColStat[23];         // List of bools deciding which columns to show in event list.

    // Log View.
    bool AutoSaveLog;                // Specifies if Explorer will save the log window at exit.

    // Work directory.
    int WorkDirType;                 // Type of working directory currently chosen. Default type = 0;
                                     // 0 = $SEISAN_TOP/WOR,
                                     // 1 = The directory that SE was launched from.
                                     // 2 = A user specified directory.
    QString UserWorkDir;             // The user spesified work directory.

    // Google Static Map.
    QString StaticMapApiKey;         // Google static map API key. Can be used with 'Show on Google Map'.
    QString DefaultMapType;          // Default map type.
    QString DefaultZoomLevel;        // Default zoom level.

    // These members that are placed here for convenience.
    // They are not saved and loaded between program sessions.
    QString AppLaunchDir;            // The directory that Explorer was launched from.
    QString SeisanWorkDir;           // The standard Seisan work directory ($SEISAN_TOP/WOR).
    QString SeisanDatDir;            // The Seisan DAT directory ($SEISAN_TOP/DAT).
};




// *****************************************************
// Structures related to the database.
// *****************************************************

enum db_type {
   default_database,                   // Database with name given in environment variable 'DEF_BASE'. Database
                                       // folders must be located in a REA folder. Location of REA folder is given
                                       // in environment variable 'SEISAN_TOP'.
   normal_database,                    // Folder containing database with year/month subfolers (C:\Seismo\REA\TEST_).
   local_database,                     // Folder with collection of S-files.
   index_file,                         // A file with list of S-files.
   no_database
};


struct db_info {
   db_type DbType;
   QString DbName;
   QString DbPath;
};


struct db_time {
   int year;
   int month;
   int day;
   int hour;
   int minute;
   float second;
};



// *****************************************************
// Structures used to hold event data in database.
// *****************************************************
struct datetime {
   short year;
   char month;
   char day;
   char hour;
   char min;
   float sec;
};

struct metadata_ {
   int event_id;                       // A unique ID for the event.
   int pos_mindex;                     // Event position in the mIndex list.
   int pos_pindex;                     // Event position in the pIndex list.
   float dist2rp;                      // Distance to ref. point. Holds value for column DstToRP.
   bool filtered_out;                  // True if event is filtered out.
   char sfile[MAX_PATH];               // Full path to S-file.
   char datetime[13];                  // Date/time for event in string format ('yyyymmddhhmm').
};

struct text_lines {
   int numlines;                       // Number of lines in the 'lines' string.
   char *lines;                        // Pointer to string with lines. All lines are 80 characters long.
};                                     // The lines in the string are NOT seperated by any character.

struct spectral_ {
   float moment;                       // Log moment, Nm
   float sdrop;                        // Stress drop, bar
   float omega0;                       // Log spectral flat level, ns
   float cornerf;                      // Corner f
   float radius;                       // Source radius
   float swin;                         // Window lenght used
   float vs;                           // S-velocity at source, km/s
   float vp;                           // P-velocity at source, km/s
   float q0;                           // Q0
   float qalpha;                       // Q alpha
   float kappa;                        // Kappa
   float density;                      // Density g/cm**3
   float slope;                        // Measured slope of spectrum
   float geo_dist;                     // Geo distance
};

struct magn_ {
   float mc;                           // Coda
   float ml;                           // Local
   float mb;                           // mb
   float ms;                           // ms
   float mw;                           // mw
};

struct phase_ {
   // Phase data parameters.
   char agency[6];                     // Agency
   char comp[5];                       // Components
   char co[3];                         // 2 letter components
   char phase[9];                      // Phase name
   char onset;                         // Onset I or E or blank
   char weight_in;                     // Input weight
   char weight_out[3];                 // Weight out
   char polarity;                      // Polarity, D or C
   int year;                           // Date and time.
   int month;
   int day;
   int hour;
   int min;
   float sec;
   double abs_time;                    // Abs time of phase time
   float coda;                         // Coda length in s
   float amp;                          // Amplitude (in Nm)
   float per;                          // Period of amplitude
   float baz_obs;                      // Observed back azimuth
   float baz_cal;                      // Calculated back azimuth
   float vel;                          // Observed apparent velocity
   float ain;                          // Observed signal to noise ratio
   float baz_res;                      // Back azimuth residual
   float res;                          // Travel time residual
   float dist;                         // Epicentral distance
   float az;                           // Azimuth
   char autoproc[21];                  // Name of auto process for making par.
   struct spectral_ spectral;          // Spectral data for phase.
   struct magn_ magnitudes;            // Magnitude data for phase.
   struct phase_ *next;                // Pointer to the next phase in linked list.
};

struct phases_ {
   int nphase;                         // Number of phases.
   struct phase_ *first;               // Pointer to first phase in list.
};

struct mag_info {
   float mag;                          // Magnitude value.
   char type;                          // Magnitude type.
   char agency[6];                     // Magnitude agency.
};

struct hyp_mag_type {
   char nmag;                          // Number of magnitudes in arrays.
   float mag[6];                       // Magnitude
   char mag_agency[6][6];              // Agency
};

struct hyp_magnitudes_ {
   struct hyp_mag_type MW;
   struct hyp_mag_type ML;
   struct hyp_mag_type MC;
   struct hyp_mag_type Mb;
   struct hyp_mag_type MB;
   struct hyp_mag_type Ms;
   struct hyp_mag_type MS;
};

struct hypocenter_ {
   struct db_time time;                // Hypocenter date and time.
   char model;                         // Model indicator
   char dist_id;                       // Distance indicator
   char type;                          // Event type like E
   char fix_org;                       // Fix origin time flag
   float lat;                          // Latitude
   float lon;                          // Longitude
   float depth;                        // Depth
   char depth_flag;                    // Depth flag
   char epi_flag;                      // Epicenter flag
   char agency[6];                     // Hypocenter agency, 3 characters only.
   int nstat;                          // Number of stations
   float rms;                          // Rms of hypocenter solution
   char high_accuracy;                 // High accurcy flag
   char error;                         // True if hypocenter has an error
   char autoproc[21];                  // Name of auto process for parameter
   float gap;                          // Gap, degrees
   float sec_err;                      // Origin time error (sec)
   float lat_err;                      // Latitude error (km)
   float lon_err;                      // intitude error (km)
   float depth_err;                    // Depth error (km)
   float cov[3];                       // Covariance, xy,xz,yz (kmXkm) (Allways 3 values here)
   struct hyp_magnitudes_ magnitudes;  // Magnitudes
   struct hypocenter_ *next;           // Pointer to the next hypocenter.
};

struct hypocenters_ {
   int nhyp;                           // Number of hypocenters.
   int nmag;                           // Number of magnitudes
   mag_info *mag_all;                  // All magnitudes. Pointer to array of mag_info structures.
   struct hypocenter_ *first;          // Pointer to first hypocenter in list.
};

struct spec_avg_ {
   float av_moment;                    // Log moment (Nm)
   float av_sdrop;                     // Stress drop, bar
   float av_omega0;                    // Log spectral flat level, ns
   float av_cornerf;                   // Corner f
   float av_radius;                    // Source radius
   float av_swin;                      // Window lenght used
   float av_mw;                        // Moment mag
   float av_slope;                     // Slope
};


struct event_node_ {
   int nstat;                          // Number of stations
   int nhead;                          // Number of header lines
   int nrecord;                        // Number of records
   int nspec;                          // Number of spectra
   char id_line[81];                   // Event ID line
   char action[4];                     // Action parameter from ID line.
   char locality[69];                  // Locality.
   struct metadata_ metadata;          // Metadata (data not read from eventfile).
   struct text_lines macros;           // Structure holding all macro data.
   struct text_lines wavefiles;        // Structure holding all wave filenames.
   struct text_lines faults;           // Structure holding all fault lines.
   struct text_lines comments;         // Structure holding all comment lines.
   struct phases_ phases;              // Structure holding all phase data.
   struct spec_avg_ spectral_avg;      // Structure holding spectral averages.
   struct hypocenters_ hypocenters;    // Structure holding all hypcocenter data.
};

#endif // TYPES_H
